/// <reference types="node" />
import { Readable } from "stream";
export declare type StringInput = string | Readable;
export interface JSONTypeSource {
    kind: "json";
    name: string;
    samples: StringInput[];
    description?: string;
}
export interface TypeScriptTypeSource {
    kind: "typescript";
    sources: {
        [filename: string]: string;
    };
}
export interface SchemaTypeSource {
    kind: "schema";
    name: string;
    uris?: string[];
    schema?: StringInput;
}
export interface GraphQLTypeSource {
    kind: "graphql";
    name: string;
    schema: any;
    query: StringInput;
}
export declare type TypeSource = GraphQLTypeSource | JSONTypeSource | SchemaTypeSource | TypeScriptTypeSource;
export declare function isJSONSource(source: TypeSource): source is JSONTypeSource;
export declare function isTypeScriptSource(source: TypeSource): source is TypeScriptTypeSource;
export declare function isSchemaSource(source: TypeSource): source is SchemaTypeSource;
export declare function isGraphQLSource(source: TypeSource): source is GraphQLTypeSource;
