"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var TypeUtils_1 = require("../TypeUtils");
var Support_1 = require("../Support");
var TypeAttributes_1 = require("../TypeAttributes");
// A union needs replacing if it contains more than one string type, one of them being
// a basic string type.
function unionNeedsReplacing(u) {
    var stringMembers = u.stringTypeMembers;
    if (stringMembers.size <= 1)
        return undefined;
    var stringType = u.findMember("string");
    if (stringType === undefined)
        return undefined;
    Support_1.assert(!TypeUtils_1.stringTypesForType(stringType).isRestricted, "We must only flatten strings if we have no restriced strings");
    return stringMembers;
}
// Replaces all string types in an enum with the basic string type.
function replaceUnion(group, builder, forwardingRef) {
    Support_1.assert(group.size === 1);
    var u = Support_1.defined(group.first());
    var stringMembers = Support_1.defined(unionNeedsReplacing(u));
    var stringAttributes = TypeUtils_1.combineTypeAttributesOfTypes("union", stringMembers);
    var types = [];
    u.members.forEach(function (t) {
        if (stringMembers.has(t))
            return;
        types.push(builder.reconstituteType(t));
    });
    if (types.length === 0) {
        return builder.getStringType(TypeAttributes_1.combineTypeAttributes("union", stringAttributes, u.getAttributes()), undefined, forwardingRef);
    }
    types.push(builder.getStringType(stringAttributes, undefined));
    return builder.getUnionType(u.getAttributes(), immutable_1.OrderedSet(types), forwardingRef);
}
function flattenStrings(graph, stringTypeMapping, debugPrintReconstitution) {
    var allUnions = graph.allNamedTypesSeparated().unions;
    var unionsToReplace = allUnions
        .filter(unionNeedsReplacing)
        .map(function (t) { return [t]; })
        .toArray();
    return graph.rewrite("flatten strings", stringTypeMapping, false, unionsToReplace, debugPrintReconstitution, replaceUnion);
}
exports.flattenStrings = flattenStrings;
